/***********************************************************************
 *
 ***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <stddef.h>

#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <sys/times.h>
#include <arpa/inet.h>

#include <epicsExport.h>
#include <devSup.h>
#include <mbboDirectRecord.h>
#include <dbAccessDefs.h>
#include <recGbl.h>
#include <errlog.h>
#include <cantProceed.h>
#include <epicsMutex.h>

#include "cdio.h"

static long init_record(mbboDirectRecord *);
static long write_mbboDirect(mbboDirectRecord *);

struct devMbboDirectCPI
{
  long          number;
  DEVSUPFUN     report;
  DEVSUPFUN     init;
  DEVSUPFUN     init_record;
  DEVSUPFUN     get_ioint_info;
  DEVSUPFUN     write_mbboDirect;
  DEVSUPFUN     special_linconv;
} devMbboDirectCPI =
  {
    6,
    NULL,
    NULL,
    init_record,
    NULL,
    write_mbboDirect,
    NULL
  };
epicsExportAddress(dset, devMbboDirectCPI);

typedef struct __cpi_mbboDirect {
  char devName[16];
  int  ch;
  int  id;
} CPI_MBBODIRECT;

extern epicsMutex lock;

/*********************************************************************
 * OUT, "@DEVNAME CH"
 *********************************************************************/
static long init_record(mbboDirectRecord *pMbboDirect)
{
  struct link *plink = &pMbboDirect->out;
  int size;
  char *buf;
  char devName[16];
  int  ch;
  long rtn = -1;

  if (plink->type != INST_IO) {
    recGblRecordError(S_db_badField,(void *)pMbboDirect,
                      "devMbboDirectCPI(init_record) Illegal INP field");
    pMbboDirect->pact = 1;
    return(S_db_badField);
  }
  
  size = strlen(plink->value.instio.string) + 1;
  buf = (char *) callocMustSucceed(size, sizeof(char), "calloc faoled");
  strncpy(buf, plink->value.instio.string, size);
  buf[size - 1] = '\0';

  if (sscanf(buf, "%s %d", devName, &ch) < 2) {
    errlogPrintf("devMbboDirectCPI(init_record): can't get device info for %s\n", pMbboDirect->name);
    pMbboDirect->pact = 1;
    return rtn;
  }

  // ch番号チェック(single end only)
  short id;
  long r = DioInit(devName, &id);
  if(r != 0) {
    errlogPrintf("devMbboDirectCPI(init_record): can't open device %s err=%ld\n", devName, r);
    pMbboDirect->pact = 1;
    return rtn;
  }
  short inPortNum, outPortNum;
  r = DioGetMaxPorts(id, &inPortNum, &outPortNum);
  if(r != 0) {
    errlogPrintf("devMbboDirectCPI(init_record): DioGetMaxCountChannels name=%s id=%d err=%ld\n", devName, id, r);
    pMbboDirect->pact = 1;
    return rtn;
  }
  if((ch < 0) || (ch > outPortNum)) {
    errlogPrintf("devMbboDirectCPI(init_record): channel number fail ch=%d name=%s id=%d err=%ld\n", ch, devName, id, r);
    pMbboDirect->pact = 1;
    return rtn;
  }
  
  CPI_MBBODIRECT *dpvt = (CPI_MBBODIRECT *)callocMustSucceed(1,
					     sizeof(CPI_MBBODIRECT),
					     "calloc failed");
  strcpy(dpvt->devName, devName);
  dpvt->ch = ch;
  dpvt->id = id;
  
  pMbboDirect->dpvt = dpvt;
  
  return 0;
}

/*********************************************************************
 *********************************************************************/
static long write_mbboDirect(mbboDirectRecord *pMbboDirect)
{
  long rtn = -1;
  CPI_MBBODIRECT dev;

  {
    epicsMutex::guard_t G(lock);
    memcpy(&dev, pMbboDirect->dpvt, sizeof(CPI_MBBODIRECT));

    short id = dev.id;

    // ビット設定
    unsigned short data = pMbboDirect->rval;
    long r = DioOutByte(id, dev.ch, data);
    if(r != 0) {
      errlogPrintf("devMbboDirectCPI(write_mbboDirect): can't get value name=%s ch=%d err=%ld\n", dev.devName, dev.ch, r);
      return rtn;
    }
  
    pMbboDirect->udf = FALSE;
  }
  return 0;
}



